use std::collections::HashMap;
use std::net::IpAddr;

use crate::constants::SECOND_FACTOR_EXEMPTION_AFTER_SUCCESSFUL_LOGIN;
use crate::data::client::ClientID;
use crate::data::entity_manager::EntityManager;
use crate::data::login_redirect::LoginRedirect;
use crate::data::totp_key::TotpKey;
use crate::data::webauthn_manager::WebauthnPubKey;
use crate::utils::err::Res;
use crate::utils::time::{fmt_time, time};

#[derive(Clone, Debug, Eq, PartialEq, serde::Serialize, serde::Deserialize)]
pub struct UserID(pub String);

#[derive(Clone, Debug, Eq, PartialEq, serde::Serialize, serde::Deserialize)]
pub struct FactorID(pub String);

#[derive(Clone, Debug, serde::Serialize, serde::Deserialize)]
pub enum TwoFactorType {
    TOTP(TotpKey),
    WEBAUTHN(Box<WebauthnPubKey>),
}

#[derive(Clone, Debug, serde::Serialize, serde::Deserialize)]
pub struct TwoFactor {
    pub id: FactorID,
    pub name: String,
    pub kind: TwoFactorType,
}

impl TwoFactor {
    pub fn type_str(&self) -> &'static str {
        match self.kind {
            TwoFactorType::TOTP(_) => "Authenticator app",
            TwoFactorType::WEBAUTHN(_) => "Security key",
        }
    }

    pub fn description_str(&self) -> &'static str {
        match self.kind {
            TwoFactorType::TOTP(_) => "Login by entering an OTP code",
            TwoFactorType::WEBAUTHN(_) => "Login using a security key",
        }
    }

    pub fn type_image(&self) -> &'static str {
        match self.kind {
            TwoFactorType::TOTP(_) => "/assets/img/pin.svg",
            TwoFactorType::WEBAUTHN(_) => "/assets/img/key.svg",
        }
    }

    pub fn login_url(&self, redirect_uri: &LoginRedirect) -> String {
        match self.kind {
            TwoFactorType::TOTP(_) => format!("/2fa_otp?redirect={}", redirect_uri.get_encoded()),
            TwoFactorType::WEBAUTHN(_) => {
                format!("/2fa_webauthn?redirect={}", redirect_uri.get_encoded())
            }
        }
    }

    pub fn is_webauthn(&self) -> bool {
        matches!(self.kind, TwoFactorType::WEBAUTHN(_))
    }
}

#[derive(Debug)]
pub struct Successful2FALogin {
    pub ip: IpAddr,
    pub time: u64,
    pub can_bypass_2fa: bool,
}

impl Successful2FALogin {
    pub fn fmt_time(&self) -> String {
        fmt_time(self.time)
    }
}

#[derive(Clone, Debug, serde::Serialize, serde::Deserialize)]
pub struct User {
    pub uid: UserID,
    pub first_name: String,
    pub last_name: String,
    pub username: String,
    pub email: String,
    pub password: String,
    pub need_reset_password: bool,
    pub enabled: bool,
    pub admin: bool,

    /// 2FA
    #[serde(default)]
    pub two_factor: Vec<TwoFactor>,

    /// Exempt the user from validating a second factor after a previous successful authentication
    /// for a defined amount of time
    #[serde(default)]
    pub two_factor_exemption_after_successful_login: bool,

    /// IP addresses of last successful logins
    #[serde(default)]
    pub last_successful_2fa: HashMap<IpAddr, u64>,

    /// None = all services
    /// Some([]) = no service
    pub authorized_clients: Option<Vec<ClientID>>,
}

impl User {
    pub fn full_name(&self) -> String {
        format!("{} {}", self.first_name, self.last_name)
    }

    pub fn quick_identity(&self) -> String {
        format!(
            "{} {} {} ({:?})",
            match self.admin {
                true => "admin",
                false => "user",
            },
            self.username,
            self.email,
            self.uid
        )
    }

    pub fn can_access_app(&self, id: &ClientID) -> bool {
        match &self.authorized_clients {
            None => true,
            Some(c) => c.contains(id),
        }
    }

    pub fn verify_password<P: AsRef<[u8]>>(&self, pass: P) -> bool {
        verify_password(pass, &self.password)
    }

    pub fn has_two_factor(&self) -> bool {
        !self.two_factor.is_empty()
    }

    pub fn can_bypass_two_factors_for_ip(&self, ip: IpAddr) -> bool {
        self.two_factor_exemption_after_successful_login
            && self.last_successful_2fa.get(&ip).unwrap_or(&0)
                + SECOND_FACTOR_EXEMPTION_AFTER_SUCCESSFUL_LOGIN
                > time()
    }

    pub fn add_factor(&mut self, factor: TwoFactor) {
        self.two_factor.push(factor);
    }

    pub fn remove_factor(&mut self, factor_id: FactorID) {
        self.two_factor.retain(|f| f.id != factor_id);
    }

    pub fn find_factor(&self, factor_id: &FactorID) -> Option<&TwoFactor> {
        self.two_factor.iter().find(|f| f.id.eq(factor_id))
    }

    pub fn has_webauthn_factor(&self) -> bool {
        self.two_factor.iter().any(TwoFactor::is_webauthn)
    }

    /// Get all registered OTP registered passwords
    pub fn get_otp_factors(&self) -> Vec<TotpKey> {
        self.two_factor.iter().fold(vec![], |mut acc, factor| {
            if let TwoFactorType::TOTP(key) = &factor.kind {
                acc.push(key.clone())
            }

            acc
        })
    }

    /// Get all registered 2FA webauthn public keys
    pub fn get_webauthn_pub_keys(&self) -> Vec<WebauthnPubKey> {
        self.two_factor.iter().fold(vec![], |mut acc, factor| {
            if let TwoFactorType::WEBAUTHN(key) = &factor.kind {
                acc.push(*key.clone())
            }

            acc
        })
    }

    /// Get the first factor of each kind of factors
    pub fn get_distinct_factors_types(&self) -> Vec<&TwoFactor> {
        let mut urls = vec![];

        self.two_factor
            .iter()
            .filter(|f| {
                if urls.contains(&f.type_str()) {
                    false
                } else {
                    urls.push(f.type_str());
                    true
                }
            })
            .collect::<Vec<_>>()
    }

    pub fn remove_outdated_successful_2fa_attempts(&mut self) {
        self.last_successful_2fa
            .retain(|_, t| *t + SECOND_FACTOR_EXEMPTION_AFTER_SUCCESSFUL_LOGIN > time());
    }

    pub fn get_formatted_2fa_successful_logins(&self) -> Vec<Successful2FALogin> {
        self.last_successful_2fa
            .iter()
            .map(|(ip, time)| Successful2FALogin {
                ip: *ip,
                time: *time,
                can_bypass_2fa: self.can_bypass_two_factors_for_ip(*ip),
            })
            .collect::<Vec<_>>()
    }
}

impl PartialEq for User {
    fn eq(&self, other: &Self) -> bool {
        self.uid.eq(&other.uid)
    }
}

impl Eq for User {}

impl Default for User {
    fn default() -> Self {
        Self {
            uid: UserID(uuid::Uuid::new_v4().to_string()),
            first_name: "".to_string(),
            last_name: "".to_string(),
            username: "".to_string(),
            email: "".to_string(),
            password: "".to_string(),
            need_reset_password: false,
            enabled: true,
            admin: false,
            two_factor: vec![],
            two_factor_exemption_after_successful_login: false,
            last_successful_2fa: Default::default(),
            authorized_clients: Some(Vec::new()),
        }
    }
}

pub fn hash_password<P: AsRef<[u8]>>(pwd: P) -> Res<String> {
    Ok(bcrypt::hash(pwd, bcrypt::DEFAULT_COST)?)
}

pub fn verify_password<P: AsRef<[u8]>>(pwd: P, hash: &str) -> bool {
    match bcrypt::verify(pwd, hash) {
        Ok(r) => r,
        Err(e) => {
            log::warn!("Failed to verify password! {:?}", e);
            false
        }
    }
}

impl EntityManager<User> {
    pub fn find_by_username_or_email(&self, u: &str) -> Option<User> {
        for entry in self.iter() {
            if entry.username.eq(u) || entry.email.eq(u) {
                return Some(entry.clone());
            }
        }
        None
    }

    pub fn find_by_user_id(&self, id: &UserID) -> Option<User> {
        for entry in self.iter() {
            if entry.uid.eq(id) {
                return Some(entry.clone());
            }
        }
        None
    }

    /// Update user information
    fn update_user<F>(&mut self, id: &UserID, update: F) -> bool
    where
        F: FnOnce(User) -> User,
    {
        let user = match self.find_by_user_id(id) {
            None => return false,
            Some(user) => user,
        };

        if let Err(e) = self.replace_entries(|u| u.uid.eq(id), &update(user)) {
            log::error!("Failed to update user information! {:?}", e);
            return false;
        }

        true
    }

    pub fn change_user_password(&mut self, id: &UserID, password: &str, temporary: bool) -> bool {
        let new_hash = match hash_password(password) {
            Ok(h) => h,
            Err(e) => {
                log::error!("Failed to hash user password! {}", e);
                return false;
            }
        };

        self.update_user(id, |mut user| {
            user.password = new_hash;
            user.need_reset_password = temporary;
            user.two_factor_exemption_after_successful_login = Default::default();
            user
        })
    }

    pub fn save_new_successful_2fa_authentication(&mut self, id: &UserID, ip: IpAddr) -> bool {
        self.update_user(id, |mut user| {
            user.last_successful_2fa.insert(ip, time());

            // Remove outdated successful attempts
            user.remove_outdated_successful_2fa_attempts();

            user
        })
    }

    pub fn clear_2fa_login_history(&mut self, id: &UserID) -> bool {
        self.update_user(id, |mut user| {
            user.last_successful_2fa = Default::default();
            user
        })
    }
}
